# -*- coding: utf-8 -*-

"""
BELA-dashboard visualizer for BELA-con version 2.x
"""

# This code is a part of belaweb system: https://github.com/letuananh/belaweb
# :developer: Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

import logging
from collections import defaultdict as dd

from bela import KNOWN_LANGUAGE_CLASSES
from bela import Bela2
from bela.lex import CorpusLexicalAnalyser

from blipleo.colors import to_color_list
from .langmixviz import LanguageTimeline
try:
    from belaweb.models import BELALexicalEntry
    __DB_AVAILABLE = True
except Exception:
    __DB_AVAILABLE = False


def build_language_color_map(language_set):
    languages = list(language_set)
    colors = to_color_list(languages, reuse_lastcolor=True)
    lang_color_map = {l: c if l not in KNOWN_LANGUAGE_CLASSES else '#999999' for l, c in zip(languages, colors)}
    return lang_color_map


def make_language_chart_json(lang_dur_map, color_map):
    language_list = [(k, v) for k, v in lang_dur_map.items()]
    language_list.sort(key=lambda x: (x[0] not in KNOWN_LANGUAGE_CLASSES, x[1]))
    langs = []
    durations = []
    for lang, dur in reversed(language_list):
        langs.append(lang)
        durations.append(round(dur, 2))
    colors = [color_map[l] for l in langs]
    return {
        'languages': langs,
        'durations': durations,
        'colors': colors,
        'all': [{'text': lang,
                 'duration': dur,
                 'color': color} for lang, dur, color in zip(langs, durations, colors)]
    }


def make_bela_language_chart_json(bela2_obj, color_map):
    ''' Calculate language chart stats '''
    lang_dur_map = dd(int)
    for tier in bela2_obj.tiers():
        if tier.tier_class == "Language":
            for ann in tier:
                if ann.duration is not None:
                    lang_dur_map[ann.value] += ann.duration
                elif ann.ref.duration is not None:
                    lang_dur_map[ann.value] += ann.ref.duration
    return make_language_chart_json(lang_dur_map, color_map)


def visualise_chatlog(bela2_obj, color_map, turn_starts, turn_takes):
    utterances = []
    max_length = 0
    for tier in bela2_obj.tiers():
        if tier.tier_class == "Utterance":
            for u in tier:
                is_error = u.errors is not None and len(u.errors) > 0
                tier_person = bela2_obj.person_map[tier.participant]
                ulength = 0
                chunks = []
                if u.chunks:
                    for child in u.chunks:
                        ulength += child.duration
                        # if not child.value or not child.value.strip():
                        #     child.errors.append("Empty chunk")
                        # elif not child.language:
                        #     if child.errors is None:
                        #         child.errors = []
                        #     child.errors.append("Language tag not found")
                        chunk_json = {
                            'text': child.value,
                            'language': child.language,
                            'color': color_map[child.language] if child.language else '',
                            'from_ts': child.from_ts.sec,
                            'to_ts': child.to_ts.sec,
                            'errors': child.errors
                        }
                        is_error = is_error or (child.errors is not None and len(child.errors) > 0)
                        chunks.append(chunk_json)
                if ulength == 0:
                    # no chunk -- probably error?
                    ulength = u.duration
                utterance_json = {
                    'text': u.value,
                    'length': ulength,
                    'chunks': chunks,
                    'speaker': tier_person.code,
                    'speaker_name': tier_person.name,
                    'is_baby': tier_person.name == "Baby",
                    'from_ts': u.from_ts.sec,
                    'errors': u.errors,
                    'is_error': is_error,
                    'is_turnA': u.ID in turn_starts,
                    'is_turnB': u.ID in turn_takes,
                    'trans': u.translation if u.translation else ''
                }
                if ulength > max_length:
                    max_length = ulength
                utterances.append(utterance_json)
    try:
        utterances.sort(key=lambda x: x['from_ts'])
    except Exception:
        if utterances:
            utterances[0]['errors'].append("Could not construct a timeline for corrupted transcription")
        pass
    speaking = ''
    for idx, chatline in enumerate(utterances):
        chatline['rduration'] = chatline['length'] * 100 / max_length
        chatline['rremainder'] = 100 - chatline['length'] * 100 / max_length
        if speaking != chatline['speaker']:
            chatline['is_continue'] = False
            speaking = chatline['speaker']
        else:
            chatline['is_continue'] = True
        for chunk in chatline['chunks']:
            try:
                chunk['rduration'] = (chunk['to_ts'] - chunk['from_ts']) * 100 / max_length
            except:
                chunk['errors'].append("Could not calculate chunk duration for corrupted annotations")
                chunk['rduration'] = 1
    return utterances


def make_people_json(bela2_obj, color_map, turn_map):
    people = []
    for person in bela2_obj.persons:
        duration = 0
        utterance_count = 0
        word_count = 0
        if person.utterances:
            for ann in person.utterances:
                duration += ann.duration
                utterance_count += 1
        lang_dur_map = dd(int)
        if person['Chunk']:
            for ann in person['Chunk']:
                if ann.words:
                    word_count += len(ann.words)
        if person['Language']:
            for ann in person['Language']:
                if ann.duration is not None:
                    lang_dur_map[ann.value] += ann.duration
                elif ann.ref is not None and ann.ref.duration is not None:
                    lang_dur_map[ann.value] += ann.ref.duration
        if utterance_count > 0:
            MLU = round(word_count / utterance_count, 3)
        else:
            MLU = 'N/A'
        person_json = {
            'name': person.name,
            'code': person.code,
            'duration': round(duration, 3),
            'languages': make_language_chart_json(lang_dur_map, color_map),
            'tiers': person.tier_classes,
            'utterance_count': utterance_count,
            'wordcount': word_count,
            'MLU': MLU,
            'turns': 0 if person.code not in turn_map else turn_map[person.code]
        }
        people.append(person_json)
    people.sort(key=lambda x: -x['utterance_count'])
    return people


def _query_custom_lang_lex_map():
    if not __DB_AVAILABLE:
        return []
    # mock-data
    # _le_map = {'English': {'meow', 'tsk'}, 'Mandarin': {'error'}}
    try:
        entries = BELALexicalEntry.objects.only('text', 'language')
    except Exception:
        logging.getLogger(__name__).exception("Could not access custom lexical entry database")
        entries = []
    _le_map = dd(set)
    for e in entries:
        _le_map[e.language].add(e.text)
    return _le_map


def analyser_lexical(bela2_obj, source='', auto_compute=True):
    analyser = CorpusLexicalAnalyser(
        lang_lex_map=_query_custom_lang_lex_map(),
        word_only=False,
        lemmatizer=True)
    for person in bela2_obj.persons:
        if person['Utterance']:
            for u in person['Utterance']:
                if u.chunks:
                    for ru in u.chunks:
                        analyser.add(ru.value, ru.language, source=source, speaker=person.code)
                else:
                    u.errors.append(f"ERROR: Utterance with no chunk")
                    analyser.add(u.value, u.language, source=source, speaker=person.code)
    if auto_compute:
        analyser.analyse()
    return analyser


def count_errors(bela2_obj):
    errors = len(bela2_obj.errors) if bela2_obj.errors else 0
    for person in bela2_obj.persons:
        if person.utterances:
            for u in person.utterances:
                if u.errors:
                    errors += len(u.errors)
    return errors


def visualize_bela2(elan, eaf_path=":memory:"):
    bela2_obj = Bela2.from_elan(elan, eaf_path=eaf_path)
    language_set = bela2_obj.get_language_set()
    color_map = build_language_color_map(language_set)
    lexicon = analyser_lexical(bela2_obj)
    lexicon_list = lexicon.to_dict()
    for l in lexicon_list:
        if 'errors' in l['stats'] and len(l['stats']['errors']):
            if bela2_obj.errors is None:
                bela2_obj.errors = []
            for e in l['stats']['errors']:
                bela2_obj.errors.append(f"Lexicon warning: {e} (Refresh in a minute to try again)")
    lexstats = {
        'stats': lexicon_list,
        'all': lexicon.profiles['ALL'].to_dict(),
        'sorting_mode': '', 'message': '', 'table_mode': True
        }
    turn_starts = set()
    turn_takes = set()
    turn_map = dd(int)
    try:
        for u, nu in bela2_obj.find_turns():
            turn_starts.add(u.ID)
            turn_takes.add(nu.ID)
            turn_map[u.person.code] += 1
    except Exception:
        bela2_obj.errors.append("Could not calculate turns")
    data = {
        'anncount': bela2_obj.count_chunks(),  # how many chunks
        'sentcount': bela2_obj.count_sents(),  # how many utterances
        'languages': make_bela_language_chart_json(bela2_obj, color_map),  # language stats
        'people': make_people_json(bela2_obj, color_map, turn_map),  # people's stats
        'chatlog': visualise_chatlog(bela2_obj, color_map, turn_starts, turn_takes),  # all utterances
        'errors': bela2_obj.errors,
        'error_count': count_errors(bela2_obj),
        'langmix_bar': LanguageTimeline.from_langmix(bela2_obj.to_language_mix().to_dict(), color_map=color_map).data,
        'lexstats': lexstats,
        'table_mode': False,
        'show_freq': False,
        'error_only': False,
        'turns_only': False,
        'belacon_version': 2.0,
        'special_lex_only': True,
        'show_translation': True
    }
    return data



# -*- coding: utf-8 -*-

"""
BELA-dashboard upload form
"""

# This code is a part of belaweb system: https://github.com/letuananh/belaweb
# :developer: Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

from django import forms


class ELANUploadForm(forms.Form):
    elanfile = forms.FileField(label="ELAN File",
                               widget=forms.FileInput(attrs={'accept': '.eaf, .bar.json, .lex.json'}))



# -*- coding: utf-8 -*-

"""
BELA-dashboard admin
"""

# This code is a part of belaweb system: https://github.com/letuananh/belaweb
# :developer: Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

from django.contrib import admin

from .models import ELANFile, BELALexicalEntry


@admin.register(ELANFile)
class ELANFileAdmin(admin.ModelAdmin):
    list_display = ('original_name', 'created_by', 'created_at', 'content', 'notes')
    readonly_fields = ('created_by', 'created_at', 'updated_by', 'updated_at', 'content')
    search_fields = ('original_name', 'notes', 'content')
    list_filter = ('created_by',)

    def save_model(self, request, obj, form, change):
        if change:
            obj.created_by = request.user
        else:
            obj.created_by = request.user
            obj.updated_by = request.user
        super().save_model(request, obj, form, change)


@admin.register(BELALexicalEntry)
class BELALexicalEntryAdmin(admin.ModelAdmin):
    list_display = ('text', 'language', 'created_by', 'updated_at')
    list_filter = ('language',)
    readonly_fields = ('created_by', 'created_at', 'updated_by', 'updated_at')

    def save_model(self, request, obj, form, change):
        if change:
            obj.created_by = request.user
        else:
            obj.created_by = request.user
            obj.updated_by = request.user
        super().save_model(request, obj, form, change)



# -*- coding: utf-8 -*-

"""
BELA-dashboard visualizer for BELA-con version 1.x
"""

# This code is a part of belaweb system: https://github.com/letuananh/belaweb
# :developer: Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

from bela import tokenize
from bela import KNOWN_LANGUAGE_CLASSES
from bela.bela1 import Bela1
from bela.bela1 import build_utterances_json
from bela.lex import CorpusLexicalAnalyser

from blipleo.colors import to_color_list
from .langmixviz import LanguageTimeline
from .bela2viz import _query_custom_lang_lex_map


def build_language_color_map(sents):
    language_set = set()
    for u in sents:
        for chunk in u:
            language_set.add(chunk['language'])
    languages = list(language_set)
    colors = to_color_list(languages, reuse_lastcolor=True)
    lang_color_map = {l: c if l not in KNOWN_LANGUAGE_CLASSES else '#999999' for l, c in zip(languages, colors)}
    return lang_color_map


def make_language_chart_json(language_list, color_map):
    ''' Transform a list of (lang,dur) into language representable list '''
    langs = []
    durations = []
    language_list.sort(key=lambda x: (x[0] not in KNOWN_LANGUAGE_CLASSES, x[1]))
    for lang, dur in reversed(language_list):
        langs.append(lang)
        durations.append(dur)
    colors = [color_map[l] for l in langs]
    return {
        'languages': langs,
        'durations': durations,
        'colors': colors,
        'all': [{'text': lang,
                 'duration': dur,
                 'color': color} for lang, dur, color in zip(langs, durations, colors)]
    }


def visualise_chatlog(utterances, color_map):
    max_length = 0
    chatlog = []
    utterance_lengths = []
    utterances.sort(key=lambda x: (x[0]['tsfrom'],))
    for u in utterances:
        ulength = 0
        for chunk in u:
            ulength += chunk['tsto'] - chunk['tsfrom']
        if ulength > max_length:
            max_length = ulength
        utterance_lengths.append(ulength)
    speaking = ''
    for idx, u in enumerate(utterances):
        chatline = {'speaker': u[0]['speaker'],
                    'speaker_name': u[0]['speaker_name'],
                    'is_continue': True,
                    'chunks': [],
                    'rduration': utterance_lengths[idx] * 100 / max_length,
                    'rremainder': 100 - utterance_lengths[idx] * 100 / max_length,
                    'is_baby': u[0]['speaker_name'] == 'Baby',
                    'errors': [],
                    'is_turnA': False,
                    'is_turnB': False,
                    'trans': ''}  # bela v1 doesn't support translation
        if speaking != u[0]['speaker']:
            chatline['is_continue'] = False
            speaking = u[0]['speaker']
        for chunk in u:
            chunk_is_error = False
            chunk_relative_duration = (chunk['tsto'] - chunk['tsfrom']) * 100 / max_length
            vchunk = {'text': chunk['text'],
                      'language': chunk['language'],
                      'rduration': chunk_relative_duration,
                      'color': color_map[chunk['language']],
                      'errors': [],
                      'is_error': chunk_is_error}
            chatline['chunks'].append(vchunk)
        chatlog.append(chatline)
    return chatlog


def make_people_json(elan_dict, utterances, color_map):
    people = elan_dict['people']
    for person in people:
        if 'languages' in person:
            person['languages'] = make_language_chart_json(person['languages'], color_map)
    people.sort(key=lambda x: -x['utterance_count'])
    # count words & MLU
    for p in people:
        p['wordcount'] = 0
        p['utterance_count'] = 0
    people_map = {p['code'][0]: p for p in people}
    for u in utterances:
        p = people_map[u[0]['speaker']]
        p['utterance_count'] += 1
        for chunk in u:
            words = tokenize(chunk['text'], language=chunk['language'])
            chunk['words'] = words
            p['wordcount'] += len(words)
    for p in people:
        p['code'] = ', '.join(p['code'])
        p['turns'] = 0
        if p['utterance_count'] > 0:
            p['MLU'] = round(p['wordcount'] / p['utterance_count'], 3)
        else:
            p['MLU'] = 'N/A'
    return people


def analyser_lexical(bela1_obj, source='', auto_compute=True):
    analyser = CorpusLexicalAnalyser(lang_lex_map=_query_custom_lang_lex_map(), word_only=False)
    for _person_name, _utterances in bela1_obj.person_utterances.items():
        for u in _utterances:
            analyser.add(u.text, u.language, source=source, speaker=_person_name)
    if auto_compute:
        analyser.analyse()
    return analyser


def visualize_bela1(elan, eaf_path=":memory:"):
    elanplus = Bela1.from_elan(elan, eaf_path=eaf_path)
    sents = build_utterances_json(elanplus)
    color_map = build_language_color_map(sents)
    elan_dict = elanplus.to_dict()
    chatlog = visualise_chatlog(sents, color_map)
    lexicon = analyser_lexical(elanplus)
    lexstats = {
        'stats': lexicon.to_dict(),
        'all': lexicon.profiles['ALL'].to_dict(),
        'sorting_mode': '', 'message': '', 'table_mode': True
        }
    data = {
        'anncount': len(elanplus.csw),  # how many chunks
        'sentcount': len(sents),  # how many utterances
        'languages': make_language_chart_json(elan_dict['languages'], color_map),  # language stats
        'people': make_people_json(elan_dict, sents, color_map),  # people's stats
        'chatlog': chatlog,  # all utterances
        'errors': [],  # TODO - support errors in PILOT1,
        'error_count': -1,
        'langmix_bar': LanguageTimeline.from_langmix(elanplus.to_language_mix().to_dict(), color_map=color_map).data,
        'lexstats': lexstats,
        'error_only': False,
        'turns_only': False,
        'table_mode': False,
        'show_freq': False,
        'belacon_version': 1.0,
        'special_lex_only': True,
        'show_translation': True
    }
    return data



# -*- coding: utf-8 -*-

"""
BELA-dashboard sunshine module
"""

# This code is a part of belaweb system: https://github.com/letuananh/belaweb
# :developer: Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

BELA_DASHBOARD_VERSION = '0.13.1'  # follow PEP-0440
BELA_DASHBOARD_BUILD = '2022-04-06'
__author__ = "Le Tuan Anh"
__email__ = "tuananh.ke@gmail.com"
__copyright__ = "Copyright (c) 2019, Le Tuan Anh <tuananh.ke@gmail.com>"
__credits__ = []
__license__ = "MIT License"
__description__ = ""
__url__ = "https://github.com/letuananh/belaweb"
__issue__ = "https://github.com/letuananh/belaweb/issues"
__source_code__ = __url__
__maintainer__ = "Le Tuan Anh"
__version_major__ = BELA_DASHBOARD_VERSION
__version__ = "{}a1".format(__version_major__)
__version_long__ = "{} - Alpha".format(__version_major__)
__status__ = "3 - Alpha"



# -*- coding: utf-8 -*-

"""
BELA-dashboard web views
"""

# This code is a part of belaweb system: https://github.com/letuananh/belaweb
# :developer: Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

import logging
from pathlib import Path

from django.contrib.auth.decorators import login_required
from django.shortcuts import render
from django.shortcuts import redirect
from django.conf import settings

from sunshine.common import log_add, log_delete
from .forms import ELANUploadForm
from .models import ELANFile
from .apps import BELA_DASHBOARD_VERSION, BELA_DASHBOARD_BUILD
from .belaviz import BELA_VERSION, BELA_DEFAULT_TITLE, SAFE_MODE
from .belaviz import BELA_DOCUMENTATION_OFFICIAL_URL
from .belaviz import BELA_DOCUMENTATION_URL
from .belaviz import BELA_DOCUMENTATION_USERNAME, BELA_DOCUMENTATION_PASSWORD
from .belaviz.common import visualize_bela, visualize_bela_lex
from .belaviz.common import visualize_langmix_bar
MEDIA_ROOT = Path(settings.MEDIA_ROOT)


def __make_bela_view_context():
    return {'title': BELA_DEFAULT_TITLE,
            'BELA_VERSION': BELA_VERSION,
            'BELA_DASHBOARD_BUILD': BELA_DASHBOARD_BUILD,
            'BELA_DASHBOARD_VERSION': BELA_DASHBOARD_VERSION,
            'BELA_DOCUMENTATION_OFFICIAL_URL': BELA_DOCUMENTATION_OFFICIAL_URL,
            'BELA_DOCUMENTATION_URL': BELA_DOCUMENTATION_URL,
            'BELA_DOCUMENTATION_USERNAME': BELA_DOCUMENTATION_USERNAME,
            'BELA_DOCUMENTATION_PASSWORD': BELA_DOCUMENTATION_PASSWORD,
            'SAFE_MODE': SAFE_MODE}


# ------------------------------------------------------------------------------
# views
# ------------------------------------------------------------------------------

@login_required()
def bela_view_file(request, module="bela", username="", filename=""):
    context = __make_bela_view_context()
    if not username or not filename:
        return render(request, 'sunshine/404.html')
    elif username != request.user.username:
        return render(request, 'sunshine/403.html')
    elanfile = ELANFile.objects.filter(created_by__username=username, content=f'{module}/{username}/{filename}')
    if not elanfile.exists():
        return render(request, 'sunshine/404.html')
    else:
        # analyse an ELAN file
        elanfile_obj = elanfile.get()
        if elanfile_obj.original_name.lower().endswith(".eaf"):
            visualize_bela(elanfile_obj, context)
            return render(request, 'belaweb/bela_visual.html', context)
        elif elanfile_obj.original_name.lower().endswith(".lex.json"):
            visualize_bela_lex(elanfile_obj, context)
            return render(request, 'belaweb/bela_lexicon.html', context)
        elif elanfile_obj.original_name.lower().endswith('.bar.json'):
            visualize_langmix_bar(elanfile_obj, context)
            return render(request, 'belaweb/bela_langbars.html', context)
        else:
            return render(request, 'belaweb/format_error.html')


@login_required
def bela_delete_file(request, module="bela", username="", filename=""):
    if not username or not filename:
        return render(request, 'sunshine/404.html')
    elif username != request.user.username:
        return render(request, 'sunshine/403.html')
    elanfile = ELANFile.objects.filter(created_by__username=username, content=f'{module}/{username}/{filename}')
    if not elanfile.exists():
        return render(request, 'sunshine/404.html')
    else:
        elanfile = elanfile.first()
        logging.getLogger(__name__).warning(f'Deleting {module}/{username}/{filename}')
        log_delete(request, elanfile,
                   change_message=f"Deleted associated file: {elanfile.content.name} | uploaded name: {elanfile.original_name}")
        elanfile.delete()
        # proceed to delete file
        return redirect(bela_home)


@login_required()
def bela_home(request):
    form = None
    context = __make_bela_view_context()
    if request.method == "POST":
        form = ELANUploadForm(request.POST, request.FILES)
        if form.is_valid():
            # try to build a model and save it ...
            elanfile = ELANFile()
            elanfile.content = form.cleaned_data['elanfile']
            elanfile.original_name = form.cleaned_data['elanfile'].name
            elanfile.created_by = request.user
            elanfile.save()
            log_add(request, elanfile)
    if form is None:
        form = ELANUploadForm()
    # list all accessible ELAN files
    context.update({
        'form': form,
        'elanfiles': ELANFile.objects.filter(created_by=request.user).order_by('-created_at')})
    return render(request, 'belaweb/home.html', context)



# -*- coding: utf-8 -*-

"""
BELA-dashboard public web config
"""

# This code is a part of belaweb system: https://github.com/letuananh/belaweb
# :developer: Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

from django.apps import AppConfig
from belaweb.belaviz import BELA_DASHBOARD_VERSION
from belaweb.belaviz import BELA_DASHBOARD_BUILD
from belaweb.belaviz import BELA_DEFAULT_TITLE
from belaweb.belaviz import BELA_DEFAULT_TITLE_SHORT
from belaweb.belaviz.common import has_access_to_bela


class BelaWebConfig(AppConfig):
    name = 'belaweb'
    verbose_name = BELA_DEFAULT_TITLE
    version = BELA_DASHBOARD_VERSION
    version_build = BELA_DASHBOARD_BUILD
    sunshine_path = 'bela/'
    namespace = ''
    sunshine_menu = [
        ('belaweb_home',
         has_access_to_bela,
         BELA_DEFAULT_TITLE_SHORT,
         'fa-scroll',
         ['Systems'])
    ]



# -*- coding: utf-8 -*-

"""
BELA-dashboard URL mapping
"""

# This code is a part of belaweb system: https://github.com/letuananh/belaweb
# :developer: Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

from django.urls import path

from . import views
from .public import views as public_views

urlpatterns = [
    path('', views.bela_home),
    path('home/', views.bela_home, name='belaweb_home'),
    path('gui/', public_views.bela_public, name='belaweb_public'),
    path('check/', views.bela_view_file, name='belaweb_visualize'),
    path('check/<slug:module>/<slug:username>/<uuid:filename>', views.bela_view_file, name='belaweb_visualize'),
    path('delete/', views.bela_delete_file, name='belaweb_deletefile'),
    path('delete/<slug:module>/<slug:username>/<uuid:filename>', views.bela_delete_file, name='belaweb_deletefile'),
]



# -*- coding: utf-8 -*-

"""
BELA-dashboard model
"""

# This code is a part of belaweb system: https://github.com/letuananh/belaweb
# :developer: Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

import logging
import uuid
from pathlib import Path
from django.conf import settings
from django.db import models
from django.contrib.auth.models import User
from django.db.models.signals import pre_delete
from django.dispatch.dispatcher import receiver


MEDIA_ROOT = Path(settings.MEDIA_ROOT)


def getLogger():
    return logging.getLogger(__name__)


def setup_filename(instance, filename):
    # suffix = Path(filename).suffix
    return f"bela/{instance.created_by.username}/{str(uuid.uuid4())}"


class ELANFile(models.Model):
    content = models.FileField(upload_to=setup_filename)
    notes = models.TextField(blank=True)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)
    original_name = models.CharField(max_length=200, blank=False)
    updated_by = models.ForeignKey(User, null=True, related_name="updated_by", on_delete=models.SET_NULL)
    created_by = models.ForeignKey(User, null=True, related_name="created_by", on_delete=models.SET_NULL)

    def __str__(self):
        return f"{self.original_name} - {self.content}"

    class Meta:
        verbose_name = 'BELA transcript'
        verbose_name_plural = 'BELA transcripts'


class BELALexicalEntry(models.Model):
    text = models.CharField(max_length=100, blank=False)
    language = models.CharField(max_length=100, blank=False)
    notes = models.TextField(blank=True)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)
    updated_by = models.ForeignKey(User, null=True, related_name="ble_updated_by", on_delete=models.SET_NULL)
    created_by = models.ForeignKey(User, null=True, related_name="ble_created_by", on_delete=models.SET_NULL)

    class Meta:
        verbose_name = 'lexical entry'
        verbose_name_plural = 'lexical entries'
        unique_together = ('text', 'language')
        indexes = [
            models.Index(fields=['text', 'language']),
        ]

    def __str__(self):
        return f"[{self.language}]/{self.text}"


@receiver(pre_delete, sender=ELANFile)
def elanfile_cleanup(sender, instance, **kwargs):
    try:
        if instance is not None and instance.content is not None:
            elanfile_path = MEDIA_ROOT / instance.content.name
            if elanfile_path.exists():
                instance.content.delete(False)
            getLogger().info(f"Deleted ELAN file: {elanfile_path}")
    except Exception:
        getLogger().exception(f"Could not clean up file {instance}")



# -*- coding: utf-8 -*-

"""
BELA-dashboard sunshine module
"""

# This code is a part of belaweb system: https://github.com/letuananh/belaweb
# :developer: Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

from .__version__ import BELA_DASHBOARD_VERSION
from .__version__ import BELA_DASHBOARD_BUILD
from .__version__ import __author__, __email__, __copyright__, __maintainer__
from .__version__ import __credits__, __license__, __description__, __url__
from .__version__ import __version_major__, __version_long__, __version__, __status__

__all__ = ['BELA_DASHBOARD_VERSION',
           'BELA_DASHBOARD_BUILD',
           "__version__", "__author__", "__description__", "__copyright__"]



# -*- coding: utf-8 -*-

"""
BELA-dashboard visualization engine package
"""

# This code is a part of belaweb system: https://github.com/letuananh/belaweb
# :developer: Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

from .bela1viz import visualize_bela1
from .bela2viz import visualize_bela2
from .common import visualize_bela
from .langmixviz import LanguageTimeline
from .common import SAFE_MODE
from bela import __version__ as BELA_VERSION
from belaweb.__version__ import BELA_DASHBOARD_VERSION, BELA_DASHBOARD_BUILD
from .common import BELA_DEFAULT_TITLE, BELA_DEFAULT_TITLE_SHORT
from .common import BELA_DOCUMENTATION_OFFICIAL_URL
from .common import BELA_DOCUMENTATION_URL
from .common import BELA_DOCUMENTATION_USERNAME
from .common import BELA_DOCUMENTATION_PASSWORD
from .common import BELA_SOURCE_REPO
from .common import BELA_ISSUE_URL
from .common import EMPTY_BELA_VIZ
from bela.lex import CorpusLexicalAnalyser

__all__ = ['visualize_bela',
           'visualize_bela1',
           'visualize_bela2',
           'LanguageTimeline',
           'BELA_VERSION', 'SAFE_MODE', 'EMPTY_BELA_VIZ',
           'BELA_DASHBOARD_VERSION', 'BELA_DASHBOARD_BUILD',
           'BELA_DEFAULT_TITLE', 'BELA_DEFAULT_TITLE_SHORT',
           'BELA_DOCUMENTATION_OFFICIAL_URL',
           'BELA_DOCUMENTATION_URL',
           'BELA_DOCUMENTATION_USERNAME',
           'BELA_DOCUMENTATION_PASSWORD',
           'BELA_SOURCE_REPO', 'BELA_ISSUE_URL',
           'CorpusLexicalAnalyser']



# Generated by Django 2.2.22 on 2021-10-20 07:34

import belaweb.models
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='ELANFile',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('content', models.FileField(upload_to=belaweb.models.setup_filename)),
                ('notes', models.TextField(blank=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('original_name', models.CharField(max_length=200)),
                ('created_by', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='created_by', to=settings.AUTH_USER_MODEL)),
                ('updated_by', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='updated_by', to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='BELALexicalEntry',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('text', models.CharField(max_length=100)),
                ('language', models.CharField(max_length=100)),
                ('notes', models.TextField(blank=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('created_by', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='ble_created_by', to=settings.AUTH_USER_MODEL)),
                ('updated_by', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='ble_updated_by', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'Lexical Entry',
                'verbose_name_plural': 'Lexical Entries',
            },
        ),
        migrations.AddIndex(
            model_name='belalexicalentry',
            index=models.Index(fields=['text', 'language'], name='belaweb_bel_text_1c0cf1_idx'),
        ),
        migrations.AlterUniqueTogether(
            name='belalexicalentry',
            unique_together={('text', 'language')},
        ),
    ]



# -*- coding: utf-8 -*-

"""
BELA-dashboard public URL mapping
"""

# This code is a part of belaweb system: https://github.com/letuananh/belaweb
# :developer: Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

from django.urls import path

from . import views

urlpatterns = [
    path('', views.bela_public),
    path('home/', views.bela_public),
    path('gui/', views.bela_public, name='belaweb_public'),
]



# -*- coding: utf-8 -*-

"""
BELA-dashboard web views
"""

# This code is a part of belaweb system: https://github.com/letuananh/belaweb
# :developer: Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

from pathlib import Path

from django.shortcuts import render
from django.conf import settings

from belaweb.forms import ELANUploadForm
from belaweb.models import ELANFile
from belaweb.belaviz import visualize_bela
from belaweb.belaviz import BELA_DOCUMENTATION_OFFICIAL_URL
from belaweb.belaviz import BELA_DEFAULT_TITLE
from belaweb.belaviz import BELA_DASHBOARD_VERSION, BELA_DASHBOARD_BUILD
from belaweb.belaviz import BELA_SOURCE_REPO, BELA_ISSUE_URL
from belaweb.belaviz import SAFE_MODE, BELA_VERSION, EMPTY_BELA_VIZ
MEDIA_ROOT = Path(settings.MEDIA_ROOT)


# ------------------------------------------------------------------------------
# views
# ------------------------------------------------------------------------------

def bela_public(request):
    context = {'title': BELA_DEFAULT_TITLE,
               'BELA_VERSION': BELA_VERSION,
               'BELA_DASHBOARD_BUILD': BELA_DASHBOARD_BUILD,
               'BELA_DASHBOARD_VERSION': BELA_DASHBOARD_VERSION,
               'BELA_DOCUMENTATION_OFFICIAL_URL': BELA_DOCUMENTATION_OFFICIAL_URL,
               'BELA_SOURCE_REPO': BELA_SOURCE_REPO,
               'BELA_ISSUE_URL': BELA_ISSUE_URL,
               'SAFE_MODE': SAFE_MODE,
               'elanfile': None,
               'data': EMPTY_BELA_VIZ}
    form = None
    if request.method == "POST":
        form = ELANUploadForm(request.POST, request.FILES)
        if form.is_valid():
            # build an ELANFile object in memory and visualise it
            # without saving to database
            elanfile_obj = ELANFile()
            elanfile_obj.content = form.cleaned_data['elanfile']
            elanfile_obj.original_name = form.cleaned_data['elanfile'].name
            visualize_bela(elanfile_obj, context)
    if form is None:
        form = ELANUploadForm()
    context['form'] = form
    return render(request, 'belaweb/bela_public.html', context)



# -*- coding: utf-8 -*-

"""
BELA visualization engine - common functions
"""

# This code is a part of belaweb system: https://github.com/letuananh/belaweb
# :developer: Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

import logging
import io
import json

from speach import elan

from django.conf import settings
from .bela1viz import visualize_bela1
from .bela2viz import visualize_bela2
from .langmixviz import LanguageTimeline

try:
    SAFE_MODE = elan.SAFE_MODE
except Exception:
    SAFE_MODE = False
from bela import __version__ as BELA_VERSION
BELA_DEFAULT_TITLE = getattr(settings, "BELA_DEFAULT_TITLE", "BELA Dashboard")
BELA_DEFAULT_TITLE_SHORT = getattr(settings, "BELA_DEFAULT_TITLE_SHORT", "BELA")
# This should be used on the public service because it points to
# stable, version-controlled copy of the BELA convention (for reproducibility of studies)
# Although the official URL is overridable in sunshine settings files
# it should not be changed without a strong reason
BELA_DOCUMENTATION_OFFICIAL_URL = getattr(settings, "BELA_DOCUMENTATION_OFFICIAL_URL", "https://blipntu.github.io/belacon/")
# This URL will be used for staff dashboard version of belaweb
# - to test out a new belacon that is not stable enough yet for the public
# - the study uses a specific (older?) version of the belacon
BELA_DOCUMENTATION_URL = getattr(settings, "BELA_DOCUMENTATION_URL", "https://blipntu.github.io/belacon/")
# Github page should not require login
BELA_DOCUMENTATION_USERNAME = getattr(settings, "BELA_DOCUMENTATION_USERNAME", "")
BELA_DOCUMENTATION_PASSWORD = getattr(settings, "BELA_DOCUMENTATION_PASSWORD", "")
BELA_SOURCE_REPO = getattr(settings, "BELA_SOURCE_REPO", "https://github.com/letuananh/belaweb")
BELA_ISSUE_URL = getattr(settings, "BELA_ISSUE_URL", "https://github.com/letuananh/belaweb/issues")
BELA_USER_GROUP = getattr(settings, "BELA_USER_GROUP", "Corpus")


def is_nested(elan):
    for tier in elan:
        if tier.children:
            return True
    return False


EMPTY_BELA_VIZ = {
    'anncount': -1,  # how many chunks
    'sentcount': -1,  # how many utterances
    'languages': {},  # language stats
    'people': {},  # people's stats
    'chatlog': {},  # all utterances
    'errors': [],
    'error_count': -1,
    'langmix_bar': {},
    'lexstats': {
        'stats': {}, 'all': {},
        'sorting_mode': '', 'message': '', 'table_mode': True},
    'table_mode': False,
    'show_freq': False,
    'error_only': False,
    'turns_only': False,
    'belacon_version': BELA_VERSION,
    'special_lex_only': False,
    'show_translation': True
}


def has_access_to_bela(req):
    try:
        return req.user.is_superuser or req.user.groups.filter(name=BELA_USER_GROUP).exists()
    except Exception:
        return False


def visualize_bela(elanfile_obj, context):
    try:
        eaf = elan.parse_eaf_stream(io.BytesIO(elanfile_obj.content.read()))
        if not is_nested(eaf):
            data = visualize_bela1(eaf, eaf_path=elanfile_obj.original_name)
        else:
            data = visualize_bela2(eaf, eaf_path=elanfile_obj.original_name)
    except Exception:
        logging.getLogger(__name__).exception(f"Could not parse BELA file {elanfile_obj.original_name} ({elanfile_obj.original_name})")
        data = EMPTY_BELA_VIZ
    __view_context = {
        'title': f'{BELA_DEFAULT_TITLE_SHORT} | {elanfile_obj.original_name}',
        'elanfile': elanfile_obj,  # database object
        'data': data
    }
    if context is not None:
        context.update(__view_context)
    return __view_context


def visualize_bela_lex(elanfile_obj, context):
    _lexstats = json.load(io.BytesIO(elanfile_obj.content.read()))
    __view_context = {
        'elanfile': elanfile_obj,
        'lexstats': _lexstats
    }
    if context is not None:
        context.update(__view_context)
    return __view_context


def visualize_langmix_bar(elanfile_obj, context):
    langmix_viz = LanguageTimeline.from_bytes(elanfile_obj.content.read())
    bars = [x.data for x in langmix_viz]
    __view_context = {
        'elanfile': elanfile_obj,
        'bar_data': bars}
    if context is not None:
        context.update(__view_context)
    return __view_context



# -*- coding: utf-8 -*-

"""
BELA-dashboard language timeline chart
"""

# This code is a part of belaweb system: https://github.com/letuananh/belaweb
# :developer: Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

import json
from collections import Counter
from blipleo.colors import to_color_list
from bela import KNOWN_LANGUAGE_CLASSES


class LanguageTimeline:
    def __init__(self, color_map=None):
        self.data = None
        self.color_map = color_map

    @property
    def chunks(self):
        return self.data

    def _process(self):
        # build language map
        languages = Counter()
        languages.update(x[0] if x[0] else '' for x in self.data['chunks'])
        if self.color_map is None:
            colors = to_color_list(languages, reuse_lastcolor=True)
            self.color_map = {l: c if l not in KNOWN_LANGUAGE_CLASSES else '#999999' for l, c in zip(languages, colors)}
        for chunk in self.data['chunks']:
            chunk_lang = chunk[0] if chunk[0] else ''
            chunk[1] = chunk[1] * 100 / self.data['length']
            chunk.append(self.color_map[chunk_lang] if chunk_lang in self.color_map else '#FF0000')
        self.data['rel_length'] = sum(chunk[1] for chunk in self.data['chunks'])
        return self.data

    @staticmethod
    def from_langmix(data, color_map=None):
        langmix_viz = LanguageTimeline(color_map=color_map)
        langmix_viz.data = data
        langmix_viz._process()
        return langmix_viz

    @staticmethod
    def from_str(text, color_map=None):
        data = json.loads(text)
        bars = []
        for lm in data:
            lm_timeline = LanguageTimeline.from_langmix(lm, color_map=color_map)
            bars.append(lm_timeline)
        return bars

    @staticmethod
    def from_bytes(b, color_map=None):
        return LanguageTimeline.from_str(b.decode('utf-8'))
    
    @staticmethod
    def read(filepath, color_map=None):
        ''' Read language timeline from a JSON file '''
        return LanguageTimeline.from_str(filepath.read_text())
